﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Newtonsoft.Json;
using Syspeace.ProviderInterface;

namespace Syspeace.Detectors.Sample.ButtonDetectorProvider
{
    public class ButtonTestRule : BaseRule {
    }

    [Serializable]
    public class ButtonTestObservation : BaseObservation {
        [ExtraInfo("Lucky number", "luckyNumber")]
        public long? LuckyNumber { get; set; }
    }

    // This is the shape of the observation JSON in the text files read.
    [JsonObject(MemberSerialization.OptOut)]
    internal class PostJsonModel {
        public string IPAddress { get; set; }
        public string UserName { get; set; }
        public bool IsSuccess { get; set; }
        public long LuckyNumber { get; set; }
        public DateTime Timestamp { get; set; }
    }

    [Detector(5)]
    public class ButtonTestDetector : IDetector<ButtonTestRule, ButtonTestObservation> {
        public ButtonTestRule[] GetDefaultRules() {
            return new [] {
                new ButtonTestRule {
                    Name = "Catch all",
                    IsEnabled = true,
                    TriggerWindow = TimeSpan.FromHours(2),
                    OccurrencesCount = 5,
                    LockoutDuration = TimeSpan.FromHours(1),
                    Priority = 1
                }, 
            };
        }

        public void UseLocation(string pathToDetectorFolder) {
        }

        class ListenerFS {
            public bool CanRunInEnvironment {
                get {
                    return Directory.Exists(basePath);
                }
            }

            private FileSystemWatcher _fileSystemWatcher;
            private IDetectorObservationListener<ButtonTestObservation> _listener;

            private const string basePath = @"C:\sp_buttonDetector";

            public void Start(IDetectorObservationListener<ButtonTestObservation> observationListener) {
                _listener = observationListener;

                var cancellationToken = _listener.CancellationToken;
                cancellationToken.Register(() => {
                    if (_fileSystemWatcher != null) {
                        _fileSystemWatcher.EnableRaisingEvents = false;
                        _fileSystemWatcher = null;
                    }
                });

                _fileSystemWatcher = new FileSystemWatcher(basePath, "*.txt");

                _fileSystemWatcher.Created += FileSystemWatcherOnCreated;
                _fileSystemWatcher.EnableRaisingEvents = true;

            }

            // TryRepeatedly is like the following code:
            // for (int i = 0; i < numTries; i++) {
            //     try {
            //         var result = insistentAction();
            //         return result;
            //     } catch (Exception) {
            //         Thread.Sleep(a small but increasing wait);
            //     }
            // }
            // except in place of Thread.Sleep, which blocks an entire thread,
            // a timer is created to run the next iteration after the delay.
            private Task<string> TryRepeatedly(Func<string> insistentAction, int numTries) {
                var taskCompletionSource = new TaskCompletionSource<string>();

                DoTryRepeatedly(insistentAction, taskCompletionSource, numTries, currentTry:0);
                
                return taskCompletionSource.Task;
            }

            private void DoTryRepeatedly(Func<string> insistentAction, TaskCompletionSource<string> taskCompletionSource, int totalTries, int currentTry) {
                try {
                    var result = insistentAction();
                    taskCompletionSource.TrySetResult(result);
                } catch (Exception) {
                    if (currentTry == totalTries) {
                        taskCompletionSource.TrySetResult(null);
                    } else {
                        Timer timer = null;
                        timer = new Timer(new TimerCallback(state => {
                            DoTryRepeatedly(insistentAction, taskCompletionSource, totalTries, currentTry + 1);
                            timer.Dispose();
                        }));
                        timer.Change(TimeSpan.FromMilliseconds(0.8*currentTry),
                            TimeSpan.FromMilliseconds(Timeout.Infinite));
                    }
                }
            }

            private void FileSystemWatcherOnCreated(object sender, FileSystemEventArgs fileSystemEventArgs) {
                switch (fileSystemEventArgs.ChangeType) {
                    case WatcherChangeTypes.Created:
                        var fullPath = fileSystemEventArgs.FullPath;

                        // Repeatedly attempt to read the file and rename it. The attempt may fail if the file is still being written, or if antivirus software is checking the file.
                        var textTask = TryRepeatedly(() => {
                            var fileText = File.ReadAllText(fullPath, Encoding.UTF8);
                            File.Move(fullPath, fullPath + "-processed");
                            return fileText;
                        }, numTries: 20);

                        textTask.ContinueWith(t => {
                            if (!t.IsCompleted || t.IsCanceled || t.IsFaulted) return;
                            var text = t.Result;

                            if (text == null) return;
                            try {
                                // Deserialize the contents of the file and assemble an observation.
                                var json = JsonConvert.DeserializeObject<PostJsonModel>(text);
                                var obs = new ButtonTestObservation {
                                    IP = json.IPAddress,
                                    LuckyNumber = json.LuckyNumber,
                                    Success = json.IsSuccess,
                                    User = json.UserName,
                                    UTCTimestamp = json.Timestamp
                                };

                                // Log and record the observation.
                                _listener.Log(string.Format("json: {0}",
                                    JsonConvert.SerializeObject(json, Formatting.None)));
                                _listener.RecordObservation(obs);
                            } catch (Exception ex) {
                                _listener.Log(string.Format("Exception: {0}", ex), important: true);
                                // Exception handling elided.
                            }
                        });

                        
                        break;
                }
            }
        }

        public void StartObserver(IDetectorObservationListener<ButtonTestObservation> observationListener) {
            _listenerFS.Start(observationListener);
        }

        public bool RuleMatchesObservation(ButtonTestRule rule, ButtonTestObservation observation) {
            return true;
        }

        public bool IsIgnorableDuplicate(ButtonTestObservation earlierObservation, ButtonTestObservation laterObservation) {
            return false;
        }

        public bool IsIgnorableOtherReason(ButtonTestObservation observation) {
            return false;
        }

        public KeyValuePair<string, string>[] FormatObservationInfo(ButtonTestObservation observation) {
            return new KeyValuePair<string, string>[] {
                new KeyValuePair<string, string>("Lucky number", observation.LuckyNumber.HasValue ? observation.LuckyNumber.Value.ToString() : "missing"), 
            };
        }

        public KeyValuePair<string, string>[] FormatRuleInfo(ButtonTestRule rule) {
            return new KeyValuePair<string, string>[0];
        }

        public string UserOrOtherOriginDescription(ButtonTestObservation observation) {
            return observation.User;
        }

        public string FormatCompositeAccountName(ButtonTestObservation observation) {
            return observation.User;
        }

        public string GetAccessReportFormattedExtra(ButtonTestObservation observation) {
            return "Lucky number " + (observation.LuckyNumber.ToString() ?? "missing");
        }

        public string FriendlyDetectorName {
            get { return "Button Test"; }
        }

        public string DetectorType {
            get { return "com.acme.ButtonTestDetector"; }
        }

        public ProviderAvailabilityResult CanRunInEnvironment {
            get { return (_listenerFS.CanRunInEnvironment ? ProviderAvailabilityResult.IsAvailable : ProviderAvailabilityResult.IsUnavailableBecause("the Button Test folder does not exist")); }
        }

        readonly ListenerFS _listenerFS = new ListenerFS();

        public string UsageWarning {
            get { return null; }
        }
    }
}
