﻿using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Http;
using System.Text;
using System.Web;
using System.Net;

namespace Syspeace.WebDetector.Reporter
{
    /// <summary>
    /// Contains methods for sending reports to Syspeace Access Log
    /// </summary>
    public static class SyspeaceReporter
    {
        const int SYSPEACE_LOG_PORT = 60100;
        /// <summary>
        /// Reports a login attempt to the Syspeace web detector
        /// </summary>
        /// <param name="username">The username attempting to login</param>
        /// <param name="success">Whether the login attempt was successful</param>
        /// <param name="request">The ASP.NET Request object</param>
        /// <param name="reportingToken">Matches the reporting token in configured Syspeace settings, if any</param>
        /// <param name="webSite">The name of the web site. May be an empty string</param>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="username"/> is null.</exception>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="request"/> is null.</exception>
        public static void LogAccess(string username, bool success, HttpRequest request, string webSite, string reportingToken = "")
        {
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }
            if (username == null)
            {
                throw new ArgumentNullException("username");
            }
            LogAccess(username, success, request.UserHostAddress, reportingToken, webSite);
        }
        /// <summary>
        /// Reports a login attempt to the Syspeace web detector
        /// </summary>
        /// <param name="username">The username attempting to login</param>
        /// <param name="success">Whether the login attempt was successful</param>
        /// <param name="baseRequest">The ASP.NET Request object</param>
        /// <param name="reportingToken">Matches the reporting token configured in Syspeace settings, if any</param>
        /// <param name="webSite">The name of the web site. May be an empty string</param>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="baseRequest"/> is null.</exception>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="username"/> is null.</exception>
        public static void LogAccess(string username, bool success, HttpRequestBase baseRequest, string webSite, string reportingToken = "")
        {
            if (baseRequest == null)
            {
                throw new ArgumentNullException("baseRequest");
            }
            if (username == null)
            {
                throw new ArgumentNullException("username");
            }
            LogAccess(username, success, baseRequest.UserHostAddress, reportingToken, webSite);
        }
        /// <summary>
        /// Reports a login attempt to the Syspeace web detector
        /// </summary>
        /// <param name="username">The username attempting to login</param>
        /// <param name="success">Whether the login attempt was successful</param>
        /// <param name="ipAddress">The IP address used by the user trying to login</param>
        /// <param name="reportingToken">Matches the Reporting Token configured in Syspeace settings, if any</param>
        /// <param name="webSite">The name of the web site. May be an empty string</param>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="username"/> is null.</exception>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="ipAddress"/> is null.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown if <paramref name="ipAddress"/> is not an IPv4 or IPv6 address.</exception>
        public static void LogAccess(string username, bool success, string ipAddress, string webSite, string reportingToken = "")
        {
            IPAddress ip;
            if (username == null)
            {
                throw new ArgumentNullException("username");
            }
            if (ipAddress == null)
            {
                throw new ArgumentNullException("ipAddress");
            }
            else if (!IPAddress.TryParse(ipAddress, out ip) || (ip.AddressFamily != System.Net.Sockets.AddressFamily.InterNetwork && ip.AddressFamily != System.Net.Sockets.AddressFamily.InterNetworkV6))
            {
                throw new ArgumentOutOfRangeException("ipAddress", "IPAddress must be an IPv4 or IPv6 address");
            }
            var jsonObj = new ReportJsonModel
            {
                UserName = username,
                Success = success,
                IP = ipAddress,
                UTCTimestamp = DateTime.UtcNow,
                WebSite = webSite,
                ReportingToken = reportingToken,

            };
            LogAccessSend(jsonObj);
        }
        /// <summary>
        /// Reports a login attempt to the Syspeace web detector
        /// </summary>
        /// <param name="username">The username attempting to login</param>
        /// <param name="success">Whether the login attempt was successful</param>
        /// <param name="ipAddress">The IP address used by the user trying to login</param>
        /// <param name="reportingToken">Matches the Reporting Token configured in Syspeace settings, if any</param>
        /// <param name="webSite">The name of the web site. May be an empty string</param>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="username"/> is null.</exception>
        /// <exception cref="System.ArgumentNullException">Thrown if <paramref name="ipAddress"/> is null.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown if <paramref name="ipAddress"/> is not an IPv4 or IPv6 address.</exception>
        public static void LogAccess(string username, bool success, IPAddress ipAddress, string webSite, string reportingToken = "")
        {
            if (username == null)
            {
                throw new ArgumentNullException("username");
            }
            if (ipAddress == null)
            {
                throw new ArgumentNullException("ipAddress");
            }
            else if ((ipAddress.AddressFamily != System.Net.Sockets.AddressFamily.InterNetwork && ipAddress.AddressFamily != System.Net.Sockets.AddressFamily.InterNetworkV6))
            {
                throw new ArgumentOutOfRangeException("ipAddress", "IPAddress must be an IPv4 or IPv6 address");
            }
            var jsonObj = new ReportJsonModel
            {
                UserName = username,
                Success = success,
                UTCTimestamp = DateTime.UtcNow,
                ReportingToken = reportingToken,
                WebSite = webSite,
            };
            LogAccessSend(jsonObj);
        }

        private static void LogAccessSend(ReportJsonModel jsonObj)
        {
            StringContent content = new StringContent(JsonConvert.SerializeObject(jsonObj));
            HttpClient client = new HttpClient();
            var result = client.PostAsync("http://localhost:" + SYSPEACE_LOG_PORT + "/report", content);
            result.ContinueWith(t =>
            {
                t.Exception.Handle(exception => true);
            }, System.Threading.Tasks.TaskContinuationOptions.OnlyOnFaulted);
        }
    }

}
